#!/bin/sh

################################################################################
# Set of LSB SpecDB consistency checks.
#
# NOTE: These checks don't duplicate LSB Navigator 'DB Consistency' section
#
# Copyright 2008 The Linux Foundation
# Copyright 2008 Institute for System Programming, RAS
#
# Author: Denis Silakov <silakov@ispras.ru>
################################################################################

dump_all=0 # Do we want to log all queries? 
	   # By default only those that found some inconsistencies are logged

if [[ $LSBDBPASSWD ]]
then
	passwd_option="-p$LSBDBPASSWD"
fi

echo "Consistency Checks Results" >consistency_result.out
echo -e "==========================\n" >>consistency_result.out

# Function to execute the query and attach its result to consistency_result.out file
call_check()
{
    arg=`echo "$@" | tr -d \\\n` # remove new lines
    query=`echo "$arg" |cut -f1 -d\;`  # query and comment are separated with ';'
    comment=`echo "$arg" |cut -f2 -d\;`
#    echo "Comment: $comment"
#    echo "Query: $query"
    
    mysql -u $LSBUSER $passwd_option $LSBDB -e "$query" |grep -v 'COUNT'> tmp_result
    
    result=`cat tmp_result`
    if [[ $result = "0" && $dump_all -eq 0 ]]
    then
	cat /dev/null >tmp_result
    fi
    
    if [[ -s tmp_result || $dump_all -ne 0 ]] 
    then
	echo -e "\n$comment" >> consistency_result.out
	for i in `seq 1 ${#comment}`
	do
	    echo -n "=" >>consistency_result.out # print necessary number of '=' under title
	done
	echo >>consistency_result.out
	cat tmp_result >> consistency_result.out
        echo >> consistency_result.out
    fi
    rm -f tmp_result
}

######################
# Queries themselves
######################

# Incorrectly included types
comment="The following Types are included, but types with such Ttype are not supposed to be included"
query="SELECT Tname,Ttype,Hname FROM Type
	LEFT JOIN HeaderGroup ON HGid=Theadgroup
        LEFT JOIN Header ON Hid=HGheader
	JOIN ArchType ON ATtid=Tid
	WHERE ATappearedin >''
	AND Ttype in ('Pointer','Const','Ref','Intrinsic','FuncPtr','Array','Unknown','Volatile') ORDER BY Hname, Ttype, Tname"
call_check "$query;$comment"

# Incorrect AIdeprecatedsince
comment="Illegal AIdeprecatedsince value"
query="SELECT AIint,AIarch,AIversion FROM ArchInt
	WHERE AIdeprecatedsince NOT LIKE '_._'"
call_check "$query;$comment"

#Incorrect Ldeprecatedsince
query="SELECT Lid FROM Library
	WHERE Ldeprecatedsince NOT LIKE '_._'"
comment="Illegal Ldeprecatedsince value"
call_check "$query;$comment"

#Incorrect SMdeprecatedsince
query="SELECT SMid FROM SubModule 
	WHERE SMdeprecatedsince NOT LIKE '_._'"
comment="Illegal SMdeprecatedsince value"
call_check "$query;$comment"

#empty Ttype - this was a popular error for some time...
query="SELECT Tname FROM Type
	WHERE Ttype=''"
comment="Empty Ttype"
call_check "$query;$comment"

# Some checks for appearedin/withdrawnin fields consistency
# Appearedin should be greater that withdrawnin - check for all tables with such fields
for table in AbiApi ArchClass ArchConst ArchDE ArchES ArchType ArchInt ArchLib \
	CmdStd Header IntStd InterpretedLanguage InterpretedLanguageModule \
	SModCmd SModLib ModSMod RpmTag TestSuite
do
	prefix=`echo $table | sed s/[[\:lower\:]]//g`
	# Ugh, RpmTag doesn't follow general naming conventions...
	if [[ $table = "RpmTag" ]]
	then
		prefix="R"
	fi
	field_appeared="${prefix}appearedin"
	field_withdrawn="${prefix}withdrawnin"
	
	query="SELECT COUNT(*) FROM $table WHERE $field_withdrawn <= $field_appeared"
	comment="Number of entries in the $table table with '$field_withdrawn' <= '$field_appeared'"
	call_check "$query;$comment"

	# Check situations when we have several records for the same element with 'withdrawnin' field equals to NULL;
	# normally this means that one of such records has empty 'appeared' field, the other one has some value in it
	# A tricky question is to guess which fields we should group by; let's take first two fields from the table
	# (with known exceptions for ArchInt and ArchClass); it would be better to use some unique key instead
	fields=`mysql -u $LSBUSER $passwd_option $LSBDB -e "show fields from $table" |cut -f1 |grep -v 'Field\|appearedin\|withdrawnin' | sed 'N;s/\n/,/' | head -1`
	if [[ $table = "ArchInt" ]]
	then
		fields="$fields,AIversion"
	elif [[ $table = "ArchClass" ]]
	then
		fields="$fields,ACpos"
	fi
	
	query="SELECT $fields FROM $table 
		WHERE $field_withdrawn IS NULL
		GROUP BY $fields
		HAVING COUNT($field_appeared) > 1"
	comment="The $table table has more than one record with $field_withdrawn = NULL for the following $fields values:"
	call_check "$query;$comment"
done

# Included types not in header
comment="Included types not assigned to header"
query="SELECT DISTINCT Tid,Tname,Ttype FROM Type 
	JOIN ArchType ON ATtid=Tid
	WHERE Ttype IN ('Typedef','Struct','Union','Enum') 
	AND Theadgroup=0 AND ATappearedin > ''"
call_check "$query;$comment"

# 'anon' types used as parameters
comment="Anon types used as parameters"
query="SELECT Pint,Ppos,Tid FROM Parameter
	LEFT JOIN Type ON Ptype=Tid
	WHERE Tname LIKE 'anon%'"
call_check "$query;$comment"
	
# 'anon' types used as return types
comment="Anon types used as return types"
query="SELECT Iid,AIreturn FROM Interface
	LEFT JOIN ArchInt ON Iid=AIint
	LEFT JOIN Type ON AIreturn=Tid
	WHERE Tname LIKE 'anon%'"
call_check "$query;$comment"

# Ilibrary should correspond to Library.Lname (through LGInt->LibGroup)
# for all symbols ever included in LSB
comment="Interface.Ilibrary not matching Library.Lname (obtained using LGInt->LibGroup link)"
query="SELECT Iname,Ilibrary FROM Interface 
	JOIN ArchInt ON AIint=Iid
	LEFT JOIN LGInt ON LGIint=Iid
	LEFT JOIN LibGroup ON LGid=LGIlibg
	LEFT JOIN Library ON Lid=LGlib
	WHERE AIappearedin > ''
	AND Lname <> Ilibrary"
call_check "$query;$comment"

# urls without baselinks
comment="Standards without baselinks, whose interfaces have non-empty urls"
query="SELECT DISTINCT Sname FROM IntStd
	LEFT JOIN Standard ON Sid=ISsid
	WHERE ISurl > ''
	AND Sbaselink IS NULL"
call_check "$query;$comment"

comment="Standards without baselinks, whose commands have non-empty urls"
query="SELECT DISTINCT Sname FROM CmdStd
	LEFT JOIN Standard ON Sid=CSsid
	WHERE CSurl > ''
	AND Sbaselink IS NULL"
call_check "$query;$comment"

echo "Foreign Keys Check" >>consistency_result.out
echo "==================" >>consistency_result.out	
cat foreign_keys | while read key_table key_field ref_table ref_field
do
        query="SELECT DISTINCT $key_field FROM $key_table
		LEFT JOIN $ref_table ON $ref_table.$ref_field=$key_table.$key_field
		WHERE $key_table.$key_field > 0 AND $ref_table.$ref_field IS NULL"
	comment="  $key_table.$key_field values not corresponding to any $ref_table.$ref_field value"
	call_check "$query;$comment"
done

query="CREATE TEMPORARY TABLE tmp1 (KEY tid(Tid)) SELECT Tid FROM Type WHERE Tid NOT IN (SELECT Ptype FROM Parameter);
	CREATE TEMPORARY TABLE tmp2 (KEY tid(Tid)) SELECT Tid FROM tmp1 WHERE Tid NOT IN (SELECT AIreturn FROM ArchInt);
	CREATE TEMPORARY TABLE tmp3 (KEY tid(Tid)) SELECT Tid FROM tmp2 WHERE Tid NOT IN (SELECT ATbasetype FROM ArchType);
	CREATE TEMPORARY TABLE tmp4 (KEY tid(Tid)) SELECT Tid FROM tmp3 WHERE Tid NOT IN (SELECT TMtypeid FROM TypeMember);
	SELECT Tid,Tname,Hname FROM tmp4
		JOIN Type using(Tid) 
		JOIN HeaderGroup ON HGid=Theadgroup 
		JOIN Header ON Hid=HGheader
		WHERE (Tname LIKE 'anon%' AND Ttype != 'Enum') OR Ttype='FuncPtr';"
comment="Unused function pointers and anonymous types (but not anon enums)"
call_check "$query;$comment"

query="SELECT DISTINCT Cname FROM Component
	WHERE Cname LIKE '%.rpm'
	OR Cname LIKE '%.deb'"
comment="Component names to be beautified"
call_check "$query;$comment"

query="SELECT DISTINCT Iclass FROM Interface
	WHERE Iclass > 0
	AND Iclass NOT IN (
	 SELECT Tid FROM Type
	 WHERE Ttype='Class'
	)"
comment="Iclass fields referencing non-existing or illegal Type entries"	

query="SELECT Tname FROM Type
	WHERE Tname LIKE '%* *%'
	OR Tname LIKE '%& &%'
	OR Tname LIKE '%& *%'
	OR Tname LIKE '%* &%'
	OR Tname LIKE '%[ %'
	OR Tname LIKE '%] [%'
	OR Tname LIKE '% ]%'
	OR Tname LIKE '%  %'
	OR Tname LIKE '%const*%'
	OR Tname LIKE '%*const%'
	OR (Ttype='FuncPtr' AND (Tname NOT LIKE '%(%' OR Tname NOT LIKE '%)%'))
	"
comment="Type names to be beautified"
call_check "$query;$comment"

query="SELECT T1.Tid, T1.Tname, ATsize, T2.Tname AS Basename FROM Type T1
	 JOIN ArchType ON ATtid=T1.Tid
	 LEFT JOIN Type T2 ON T2.Tid=ATbasetype
	 WHERE T1.Ttype='Array'
	 AND T1.Tname <> CONCAT(T2.Tname,'[', ATsize, ']')
	 AND T1.Tname <> CONCAT(T2.Tname,'[...]')
	 AND (T2.Tunmangled IS NULL OR T1.Tname <> CONCAT(T2.Tunmangled,'[', ATsize, ']'))
	 AND T2.Ttype<>'FuncPtr'
	 "
comment="Type names of arrays to be beautified or corrected"
call_check "$query;$comment"
